@@ -38,6 +38,7 @@ from datetime import datetime
 
 from txaio.interfaces import IFailedFuture, ILogger, log_levels
 from txaio._iotype import guess_stream_needs_encoding
+from txaio._common import _BatchedTimer
 from txaio import _Config
 
 import six
@@ -54,22 +55,22 @@ except ImportError:
     from trollius import iscoroutine
     from trollius import Future
 
-    class PrintHandler(logging.Handler):
-        def emit(self, record):
-            print(record)
-    logging.getLogger("trollius").addHandler(PrintHandler())
-
-
 config = _Config()
 config.loop = asyncio.get_event_loop()
 _stderr, _stdout = sys.stderr, sys.stdout
-_loggers = []  # weak-ref's of each logger we've created before start_logging()
+_loggers = weakref.WeakSet()  # weak-ref's of each logger we've created before start_logging()
 _log_level = 'info'  # re-set by start_logging
+_started_logging = False
+_categories = {}
 
 using_twisted = False
 using_asyncio = True
 
 
+def add_log_categories(categories):
+    _categories.update(categories)
+
+
 class FailedFuture(IFailedFuture):
     """
     This provides an object with any features from Twisted's Failure
@@ -102,16 +103,24 @@ class FailedFuture(IFailedFuture):
 
 # API methods for txaio, exported via the top-level __init__.py
 
-def _log(logger, level, msg, **kwargs):
+def _log(logger, level, format=u'', **kwargs):
+
+    # Look for a log_category, switch it in if we have it
+    if "log_category" in kwargs and kwargs["log_category"] in _categories:
+        format = _categories.get(kwargs["log_category"])
+
     kwargs['log_time'] = time.time()
     kwargs['log_level'] = level
-    kwargs['log_message'] = msg
+    kwargs['log_format'] = format
     # NOTE: turning kwargs into a single "argument which
     # is a dict" on purpose, since a LogRecord only keeps
     # args, not kwargs.
     if level == 'trace':
         level = 'debug'
-    getattr(logger._logger, level)(msg, kwargs)
+        kwargs['txaio_trace'] = True
+
+    msg = format.format(**kwargs)
+    getattr(logger._logger, level)(msg)
 
 
 def _no_op(*args, **kw):
@@ -121,18 +130,23 @@ def _no_op(*args, **kw):
 class _TxaioLogWrapper(ILogger):
     def __init__(self, logger):
         self._logger = logger
-        self._set_level(_log_level)
+        self._set_log_level(_log_level)
 
-    def _set_level(self, level):
+    def emit(self, level, *args, **kwargs):
+        func = getattr(self, level)
+        return func(*args, **kwargs)
+
+    def _set_log_level(self, level):
         target_level = log_levels.index(level)
         # this binds either _log or _no_op above to this instance,
         # depending on the desired level.
         for (idx, name) in enumerate(log_levels):
-            if idx < target_level:
+            if idx <= target_level:
                 log_method = functools.partial(_log, self, name)
             else:
                 log_method = _no_op
             setattr(self, name, log_method)
+        self._log_level = level
 
 
 class _TxaioFileHandler(logging.Handler, object):
@@ -142,11 +156,19 @@ class _TxaioFileHandler(logging.Handler, object):
         self._encode = guess_stream_needs_encoding(fileobj)
 
     def emit(self, record):
-        fmt = record.args['log_message']
-        dt = datetime.fromtimestamp(record.args['log_time'])
+        if isinstance(record.args, dict):
+            fmt = record.args.get(
+                'log_format',
+                record.args.get('log_message', u'')
+            )
+            message = fmt.format(**record.args)
+            dt = datetime.fromtimestamp(record.args.get('log_time', 0))
+        else:
+            message = record.getMessage()
+            dt = datetime.fromtimestamp(record.created)
         msg = u'{0} {1}{2}'.format(
             dt.strftime("%Y-%m-%dT%H:%M:%S%z"),
-            fmt.format(**record.args),
+            message,
             os.linesep
         )
         if self._encode:
@@ -157,20 +179,20 @@ class _TxaioFileHandler(logging.Handler, object):
 def make_logger():
     logger = _TxaioLogWrapper(logging.getLogger())
     # remember this so we can set their levels properly once
-    # start_logging is actually called.
-    if _loggers is not None:
-        _loggers.append(weakref.ref(logger))
+    # start_logging is actually called
+    _loggers.add(logger)
     return logger
 
 
-def start_logging(out=None, level='info'):
+def start_logging(out=_stdout, level='info'):
     """
     Begin logging.
 
-    :param out: if provided, a file-like object to log to
+    :param out: if provided, a file-like object to log to. By default, this is
+                stdout.
     :param level: the maximum log-level to emit (a string)
     """
-    global _log_level, _loggers
+    global _log_level, _loggers, _started_logging
     if level not in log_levels:
         raise RuntimeError(
             "Invalid log level '{0}'; valid are: {1}".format(
@@ -178,12 +200,12 @@ def start_logging(out=None, level='info'):
             )
         )
 
-    if _loggers is None:
+    if _started_logging:
         return
+
+    _started_logging = True
     _log_level = level
 
-    if out is None:
-        out = _stdout
     handler = _TxaioFileHandler(out)
     logging.getLogger().addHandler(handler)
     # note: Don't need to call basicConfig() or similar, because we've
@@ -200,11 +222,8 @@ def start_logging(out=None, level='info'):
     logging.getLogger().setLevel(level_to_stdlib[level])
     # make sure any loggers we created before now have their log-level
     # set (any created after now will get it from _log_level
-    for ref in _loggers:
-        instance = ref()
-        if instance is not None:
-            instance._set_level(level)
-    _loggers = None
+    for logger in _loggers:
+        logger._set_log_level(level)
 
 
 def failure_message(fail):
@@ -296,6 +315,27 @@ def call_later(delay, fun, *args, **kwargs):
     return config.loop.call_later(delay, real_call)
 
 
+def make_batched_timer(bucket_seconds, chunk_size=100):
+    """
+    Creates and returns an object implementing
+    :class:`txaio.IBatchedTimer`.
+
+    :param bucket_seconds: the number of seconds in each bucket. That
+        is, a value of 5 means that any timeout within a 5 second window
+        will be in the same bucket, and get notified at the same time.
+
+    :param chunk_size: when "doing" the callbacks in a particular
+        bucket, this controls how many we do at once before yielding to
+        the reactor.
+    """
+    get_seconds = config.loop.time
+    return _BatchedTimer(
+        bucket_seconds, chunk_size,
+        seconds_provider=get_seconds,
+        delayed_call_creator=call_later,
+    )
+
+
 def is_called(future):
     return future.done()
 
@@ -366,3 +406,17 @@ def gather(futures, consume_exceptions=True):
     # exception will be immediately propagated to the returned
     # future."
     return asyncio.gather(*futures, return_exceptions=consume_exceptions)
+
+
+def set_global_log_level(level):
+    """
+    Set the global log level on all loggers instantiated by txaio.
+    """
+    for logger in _loggers:
+        logger._set_log_level(level)
+    global _log_level
+    _log_level = level
+
+
+def get_global_log_level():
+    return _log_level
